#ifndef VKTOOLS_H
#define VKTOOLS_H

#include "vktool.h"

CREATE_VKTOOL(CropFromFiducials)
void CropFromFiducials::initialize(void)
{
    createMethod("CropFromFiducials", "Crops all images. Fiducial points define borders.");
}
bool CropFromFiducials::runMethod()
{
    unsigned int i = 0;
    Byte3DImage::IndexType f;
    Byte3DImage::IndexType min, max;

    if(getFiducial(i, &f))
    min = f; max = f;
    i++;
    while(getFiducial(i, &f))
    {
        for(unsigned int d = 0; d < 3; d++)
        {
            if(min[d] > f[d]) min[d] = f[d];
            if(max[d] < f[d]) max[d] = f[d];
        }
        i++;
    }
//    for(unsigned int d = 0; d < 3; d++)
//    {
//        if(min[d] > max[d]) return false;
//    }
    return crop(min, max);
}


#include <itkInvertIntensityImageFilter.h>
CREATE_VKTOOL(Inverter)
void Inverter::initialize(void)
{
    createMethod("Inversion", "Computes negative - inverts image gray scale");
}
bool Inverter::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    typedef itk::InvertIntensityImageFilter <Byte3DImage> InvertIntensityImageFilterType;
    InvertIntensityImageFilterType::Pointer invertIntensityFilter = InvertIntensityImageFilterType::New();
    invertIntensityFilter->SetInput(mi);
    invertIntensityFilter->SetMaximum(255);
    invertIntensityFilter->Update();
    Byte3DImage::Pointer output = invertIntensityFilter->GetOutput();
    ImageWithInfo oui;
    oui.i = output;
    oui.t = name;
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(ColorMultiplier)
void ColorMultiplier::initialize(void)
{
    createMethod("Color multiplication", "Multiplies RGB channel images");
    createParameter("GammaRed", "f;1.0");
    createParameter("GammaGreen", "f;1.0");
    createParameter("GammaBlue", "f;1.0");
    createParameter("Multiplier", "f;1.0");
}
bool ColorMultiplier::runMethod(void)
{
    unsigned int d;
    float weightRed = getFloatParameter("GammaRed");
    float weightGreen = getFloatParameter("GammaGreen");
    float weightBlue = getFloatParameter("GammaBlue");
    float bias = getFloatParameter("Multiplier");

    std::string redName, blueName, greenName;
    ImageWithInfo* inputRedI = getRedImage(&redName);
    ImageWithInfo* inputGreenI = getGreenImage(&greenName);
    ImageWithInfo* inputBlueI = getBlueImage(&blueName);

    if(inputRedI == NULL || inputBlueI == NULL || inputGreenI == NULL) return false;
    Byte3DImage::Pointer inputRed = inputRedI->i;
    Byte3DImage::Pointer inputGreen = inputGreenI->i;
    Byte3DImage::Pointer inputBlue = inputBlueI->i;

    typedef itk::ImageDuplicator< Byte3DImage > DuplicatorType;
    DuplicatorType::Pointer duplicator = DuplicatorType::New();
    duplicator->SetInputImage(inputRed);
    duplicator->Update();
    Byte3DImage::Pointer output = duplicator->GetOutput();

    size_t max = 1;
    for(d = 0; d < 3; d++) max *= output->GetLargestPossibleRegion().GetSize()[d];

    unsigned char* ptrr = inputRed->GetPixelContainer()->GetImportPointer();
    unsigned char* ptrg = inputGreen->GetPixelContainer()->GetImportPointer();
    unsigned char* ptrb = inputBlue->GetPixelContainer()->GetImportPointer();
    unsigned char* ptro = output->GetPixelContainer()->GetImportPointer();
    for(d = 0; d < max; d++)
    {
        double v =  pow((double)*ptrr/255.0, weightRed) * pow((double)*ptrg/255.0, weightGreen) * pow((double)*ptrb/255.0, weightBlue) * bias * 255.0;
        if(v > 255) v = 255;
        if(v < 0) v = 0;
        *ptro = v;
        ptro++; ptrr++; ptrg++; ptrb++;
    }
    ImageWithInfo oui;
    std::stringstream ss;
    ss << name << " "<< bias << "*" << redName << "^" << weightRed << "*" << greenName << "^" << weightGreen << "+" << blueName << "^" <<  weightBlue;
    oui.t = ss.str();
    oui.i = output;
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(ColorAddition)
void ColorAddition::initialize(void)
{
    createMethod("Color addition", "Combines RGB channel images");
    createParameter("WeightRed", "f;1.0");
    createParameter("WeightGreen", "f;1.0");
    createParameter("WeightBlue", "f;1.0");
    createParameter("Bias", "f;0.0");
}
bool ColorAddition::runMethod(void)
{
    unsigned int d;
    float weightRed = getFloatParameter("WeightRed");
    float weightGreen = getFloatParameter("WeightGreen");
    float weightBlue = getFloatParameter("WeightBlue");
    float bias = getFloatParameter("Bias");

    std::string redName, blueName, greenName;
    ImageWithInfo* inputRedI = getRedImage(&redName);
    ImageWithInfo* inputGreenI = getGreenImage(&greenName);
    ImageWithInfo* inputBlueI = getBlueImage(&blueName);

    if(inputRedI == NULL || inputBlueI == NULL || inputGreenI == NULL) return false;
    Byte3DImage::Pointer inputRed = inputRedI->i;
    Byte3DImage::Pointer inputGreen = inputGreenI->i;
    Byte3DImage::Pointer inputBlue = inputBlueI->i;

    typedef itk::ImageDuplicator< Byte3DImage > DuplicatorType;
    DuplicatorType::Pointer duplicator = DuplicatorType::New();
    duplicator->SetInputImage(inputRed);
    duplicator->Update();
    Byte3DImage::Pointer output = duplicator->GetOutput();

    size_t max = 1;
    for(d = 0; d < 3; d++) max *= output->GetLargestPossibleRegion().GetSize()[d];

    unsigned char* ptrr = inputRed->GetPixelContainer()->GetImportPointer();
    unsigned char* ptrg = inputGreen->GetPixelContainer()->GetImportPointer();
    unsigned char* ptrb = inputBlue->GetPixelContainer()->GetImportPointer();
    unsigned char* ptro = output->GetPixelContainer()->GetImportPointer();
    for(d = 0; d < max; d++)
    {
        float v = weightRed**ptrr + weightGreen**ptrg + weightBlue**ptrb + bias;
        if(v > 255) v = 255;
        if(v < 0) v = 0;
        *ptro = v;
        ptro++; ptrr++; ptrg++; ptrb++;
    }
    ImageWithInfo oui;
    std::stringstream ss;
    ss << name << " " << weightRed << "*" << redName << "+" << weightGreen << "*" << greenName << "+" << weightBlue << "*" << blueName << "+" << bias;
    oui.t = ss.str();
    oui.i = output;
    appendImage(oui, name);
    return true;
}

#include "../shared/mz2itk.h"
#include "../shared/imagefilters.h"
#include "../shared/imagefilters2.h"
CREATE_VKTOOL(GaussianLowPass)
void GaussianLowPass::initialize(void)
{
    createMethod("Low-pass Gaussian filtration", "Image convolution with Gaussian kernel");
    createParameter("Sigma", "f;1.0;0.0");
}
bool GaussianLowPass::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    float sigma = getFloatParameter("Sigma");
    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters<float>::gaussianFilter(ii, sigma);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Sigma = " << sigma;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(Gradient)
void Gradient::initialize(void)
{
    createMethod("Gradient magnitude", "Compute magnitude of image gradient");
}
bool Gradient::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters<float>::gradientFilter(ii);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    oui.t = name;
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(Laplacian)
void Laplacian::initialize(void)
{
    createMethod("Laplacian", "Compute Laplacian");
}
bool Laplacian::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters<float>::laplacianFilter(ii);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    oui.t = name;
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(CannyEdge)
void CannyEdge::initialize(void)
{
    createMethod("Canny edges", "Canny edge detection");
    createParameter("Sigma", "f;1.0;0.0");
    createParameter("LowerThreshold", "i;1;0;255");
    createParameter("UpperThreshold", "i;254;0;255");
}
bool CannyEdge::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;

    float sigma = getFloatParameter("Sigma");
    int min = getIntParameter("LowerThreshold");
    int max = getIntParameter("UpperThreshold");
    Byte3DImage::Pointer mi = mmi->i;
    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters<float>::cannyEdgeFilter(ii, sigma, min, max);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Sigma = " << sigma << ", Lower = " << min << ", Upper = " << max;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}

/*
    static typename ImageType::Pointer cannyEdgeFilter(typename ImageType::Pointer input_image, float sigma, float upperThreshold, float lowerThreshold)
*/

CREATE_VKTOOL(Threshold)
void Threshold::initialize(void)
{
    createMethod("Thresholding", "Image binarization by gray-scale thresholding");
    createParameter("Min", "i;128;0;255");
    createParameter("Max", "i;255;0;255");
}
bool Threshold::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;
    int min = getIntParameter("Min");
    int max = getIntParameter("Max");

    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    Mz2Itk<unsigned char>::Image3DType::Pointer oi = ImageFilters<unsigned char>::thresholdFilter(ii, min, max);
    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZNormalized(oi);

    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Min = " << min << ", Max = " << max;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(FloodFillO)
void FloodFillO::initialize(void)
{
    createMethod("Flood fill", "Region growing - bucket fill tool");
    createParameter("Fiducials", "t;current;current;all");
    createParameter("Min", "i;128;0;255");
    createParameter("Max", "i;255;0;255");
}
bool FloodFillO::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    int min = getIntParameter("Min");
    int max = getIntParameter("Max");
    std::string mode = getStringParameter("Fiducials");


    Byte3DImage::IndexType fiducial;


   // if(mode == "all")


    if(!getCurrentFiducial(&fiducial)) return false;
    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    Mz2Itk<unsigned char>::Image3DType::Pointer oi = ImageFilters<unsigned char>::regionGrowing(ii, min, max, fiducial);
    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZNormalized(oi);

    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Min = " << min << ", Max = " << max;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(Median)
void Median::initialize(void)
{
    createMethod("Median filtration", "Median filtration");
    createParameter("Radius", "f;1.0;0.0");
}
bool Median::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;
    float radius = getFloatParameter("Radius");
    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    ImageFilters<unsigned char>::ImageType::Pointer oi = ImageFilters<unsigned char>::median(ii, radius);
    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Radius = " << radius;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(Dilation)
void Dilation::initialize(void)
{
    createMethod("Dilation", "Morphological dilation");
    createParameter("Radius", "f;1.0;0.0");
}
bool Dilation::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;
    float radius = getFloatParameter("Radius");
    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    ImageFilters<unsigned char>::ImageType::Pointer oi = ImageFilters<unsigned char>::dilate(ii, radius);
    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Radius = " << radius;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(Erosion)
void Erosion::initialize(void)
{
    createMethod("Erosion", "Morphological erosion");
    createParameter("Radius", "f;1.0;0.0");
}
bool Erosion::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;
    float radius = getFloatParameter("Radius");
    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    ImageFilters<unsigned char>::ImageType::Pointer oi = ImageFilters<unsigned char>::erode(ii, radius);
    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Radius = " << radius;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}


#include <itkHessianRecursiveGaussianImageFilter.h>
#include "../shared/itkScaffoldHessianToVesselnessFilter.h"
CREATE_VKTOOL(Vesselness)
void Vesselness::initialize(void)
{
    createMethod("Vesselness", "Hessian based vesselness function");
    createParameter("Method", "t;Frangi;Sato;Frangi;Erdt;FrangiXY;SatoXY;ErdtXY");
    createParameter("NumberOfScales", "i;1;1");
    createParameter("StdevMin", "f;1.0;0.0", "Standard deviation for minimum scale");
    createParameter("StdevMax", "f;1.0;0.0", "Standard deviation for maximum scale");
    createParameter("Enhancement", "f;1.0;0.0;10.0", "Enhancement of higher scales");
    createParameter("C", "f;1.0;0.0", "Frangi's algorithm C parameter");
    createParameter("Alpha", "f;0.5;0.0", "Frangi's algorithm Alpha parameter");
    createParameter("Beta", "f;0.5;0.0", "Frangi's algorithm Beta parameter");
    createParameter("Alpha1", "f;0.5;0.0", "Sato's algorithm Alpha1 parameter");
    createParameter("Alpha2", "f;2.0;0.0", "Sato's algorithm Alpha2 parameter");
}
bool Vesselness::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    std::string method = getStringParameter("Method");
    int scales = getIntParameter("NumberOfScales");
    float stdmin = getFloatParameter("StdevMin");
    float stdmax = getFloatParameter("StdevMax");
    float enhanc = getFloatParameter("Enhancement");
    float c = getFloatParameter("C");
    float alpha = getFloatParameter("Alpha");
    float beta = getFloatParameter("Beta");
    float alpha1 = getFloatParameter("Alpha1");
    float alpha2 = getFloatParameter("Alpha2");

    typedef itk::HessianRecursiveGaussianImageFilter< Mz2Itk<float>::Image3DType > F1;
    typedef F1::Pointer P1;
    typedef itk::ScaffoldHessianToVesselnessFilter< float > F2;
    typedef F2::Pointer P2;
    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    typedef itk::ImageRegionConstIterator< Mz2Itk<float>::Image3DType > ConstIteratorType;
    typedef itk::ImageRegionIterator< Mz2Itk<float>::Image3DType > IteratorType;


    P1 filterH = F1::New();
    filterH->SetInput(ii);
    P2 filterV = F2::New();

    if(method == "Sato") filterV->SetMethodSato(alpha1, alpha2);
    else if(method == "Erdt") filterV->SetMethodErdt();
    else if(method == "FrangiXY") filterV->SetMethodFrangiXY(c, alpha, beta);
    else if(method == "SatoXY") filterV->SetMethodSatoXY(alpha1, alpha2);
    else if(method == "ErdtXY") filterV->SetMethodErdtXY();
    else filterV->SetMethodFrangi(c, alpha, beta);

    filterH->SetSigma(stdmin*stdmin);
    filterH->Update();
    filterV->SetInput(filterH->GetOutput());
    filterV->Update();
    Mz2Itk<float>::Image3DType::Pointer result = filterV->GetOutput();

    double q = pow((double)stdmax/stdmin,(1.0/double(scales-1)));
    for(int s = 1; s < scales; s++)
    {
        float stdev = stdmin*pow(q, (double)s);
        filterH->SetSigma(stdev*stdev);
        filterH->Update();
        filterV->SetInput(filterH->GetOutput());
        filterV->Update();
        Mz2Itk<float>::Image3DType::Pointer res = filterV->GetOutput();
        double multip = stdev/stdmin;
        multip = pow(multip, enhanc);

        IteratorType firstIterator(result, result->GetRequestedRegion());
        ConstIteratorType nextIterator(res, res->GetRequestedRegion());
        firstIterator.GoToBegin();
        nextIterator.GoToBegin();
        while (!firstIterator.IsAtEnd())
        {
            if (multip*nextIterator.Get() > firstIterator.Get())
                firstIterator.Set(multip*nextIterator.Get());
            ++firstIterator;
            ++nextIterator;
        }
    }
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(result);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Method: "<< method <<", Scales = " << scales << ": <" << stdmin << ", " << stdmax;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(HoleFill)
void HoleFill::initialize(void)
{
    createMethod("HoleFill", "Fills spaces in binary objects");
    createParameter("Object", "t;White;White;Black");

}
bool HoleFill::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    std::string object = getStringParameter("Object");

    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    ImageFilters< unsigned char >::ImageType::Pointer oi;
    unsigned char foreground;
    if(object == "Black")
    {
        foreground = ImageFilters< unsigned char >::minIntensity(ii);
    }
    else
    {
        foreground = ImageFilters< unsigned char >::maxIntensity(ii);
    }
    oi = ImageFilters<unsigned char>::holeFill(ii, foreground);

    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZ(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(TopHat)
void TopHat::initialize(void)
{
    createMethod("TopHat", "Top hat transform");
    createParameter("Object", "t;White;White;Black");
    createParameter("Radius", "f;1.0;0.0");
}
bool TopHat::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    std::string object = getStringParameter("Object");
    float radius = getFloatParameter("Radius");
    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    ImageFilters<float>::ImageType::Pointer oi;
    if(object == "Black") oi = ImageFilters<float>::blackTopHat(ii, radius);
    else oi = ImageFilters<float>::whiteTopHat(ii, radius);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name << ", Radius = " << radius;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(SkeletonFromBinary)
void SkeletonFromBinary::initialize(void)
{
    createMethod("SkeletonFromBinary", "Fills spaces in binary images");
}
bool SkeletonFromBinary::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;
    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    Mz2Itk<unsigned char>::Image3DType::Pointer oi = ImageFilters<unsigned char>::skeletonFromBinary(ii);
    Byte3DImage::Pointer ni = Mz2Itk<unsigned char>::getMZNormalized(oi);
    ImageWithInfo oui;
    oui.i = ni;
    std::stringstream ss;
    ss << name;
    oui.t = ss.str();
    appendImage(oui, name);
    return true;
}

CREATE_VKTOOL(TreeFromSkeleton)
void TreeFromSkeleton::initialize(void)
{
    createMethod("TreeFromSkeleton", "Center lines vectorization");
}
bool TreeFromSkeleton::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage();
    if(mmi == NULL) return false;
    Byte3DImage::Pointer mi = mmi->i;
    Mz2Itk<unsigned char>::Image3DType::Pointer ii = Mz2Itk<unsigned char>::getItk(mi);
    Tree tree;
    Byte3DImage::IndexType fiducial;
    if(getCurrentFiducial(&fiducial)) tree = BuildTree<unsigned char>::skeletonToTree(ii, &fiducial);
    else tree = BuildTree<unsigned char>::skeletonToTree(ii, NULL);
    appendTree(tree, name);
    return true;
}


CREATE_VKTOOL(TreeFromUpsizedBinary)
void TreeFromUpsizedBinary::initialize(void)
{
    createMethod("TreeFromUpsizedBinary", "");
    createParameter("UpsizeMethod", "t;NeighborCheck;NeighborCheck;ItkResampleImageFilter");
}
bool TreeFromUpsizedBinary::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;
    std::string mode = getStringParameter("UpsizeMethod");
    Mz2Itk<unsigned char>::Image3DType::Pointer ii;
    if(mode == "ItkResampleImageFilter")  ii = Mz2Itk<unsigned char>::getUpsizedBinaryI(mi);
    else ii = Mz2Itk<unsigned char>::getUpsizedBinaryM(mi);
    Mz2Itk<unsigned char>::Image3DType::Pointer oi = ImageFilters<unsigned char>::skeletonFromBinary(ii);
    Tree tree;
    Byte3DImage::IndexType fiducial;
    if(getCurrentFiducial(&fiducial)) tree = BuildTree<unsigned char>::skeletonToTree(oi, &fiducial);
    else tree = BuildTree<unsigned char>::skeletonToTree(oi, NULL);
    appendTree(tree, name);
    return true;
}


#include "../shared/treediameters.h"
CREATE_VKTOOL(TreeDiametersFromBinary)
void TreeDiametersFromBinary::initialize(void)
{
    createMethod("TreeDiametersFromBinary", "Estimates diameters of existing tree from binary image");
    createParameter("Density", "i;2;1", "Divider for icosahedron based mesh");
    createParameter("Gamma", "f;4.0;0.0", "Weights impact of smallest to second smallest eigenvalue");
    createParameter("Rho", "f;1.4142136;0.0", "Multiplier, should be equal to 1.4142");
}
bool TreeDiametersFromBinary::runMethod()
{
    Tree* mt = getCurrentTree();
    if(mt == 0) return false;
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;

    int divider = getIntParameter("Density");
    float gamma = getFloatParameter("Gamma");
    float rho = getFloatParameter("Rho");

    TreeDiameters diams(mt, mi, gamma, divider, rho);
    Tree tree = diams.getResult();

    appendTree(tree, name);
    return true;
}


#include "../shared/treesmooth.h"
CREATE_VKTOOL(SmoothCenterlines)
void SmoothCenterlines::initialize(void)
{
    createMethod("SmoothTreeBranches", "Smoothes centerlines");
    createParameter("Alpha", "f;0.5;0.0", "Controls minimization of the first derivative");
    createParameter("Beta", "f;0.2;0.0", "Controls minimization of the second derivative");
    createParameter("Spring", "f;0.1;0.0", "Minimizes distance from original location");
    createParameter("Anchor", "f;0.1;0.0", "Minimizes third power of distance from original location");
    createParameter("Iterations", "i;5;1", "Number of iterations");
    createParameter("Chillout", "i;5;0", "Number of finishing iterations");
}
bool SmoothCenterlines::runMethod()
{
    Tree* mt = getCurrentTree();
    if(mt == 0) return false;
    float d1 = getFloatParameter("Alpha");;
    float d2 = getFloatParameter("Beta");;
    float fa = getFloatParameter("Spring");;
    float faaa = getFloatParameter("Anchor");;
    int iterations = getIntParameter("Iterations");;
    int chillout = getIntParameter("Chillout");;

    TreeSmooth smoothing(mt, d1, d2, fa, faaa);
    Tree tree = smoothing.getResult(iterations, chillout);

    appendTree(tree, name);
    return true;
}

#include "../shared/smoothdiameter.h"
CREATE_VKTOOL(SmoothDiameters)
void SmoothDiameters::initialize(void)
{
    createMethod("SmoothDiameters", "Smoothes diameters");
    createParameter("Alpha", "f;0.5;0.0", "Controls minimization of the first derivative");
    createParameter("Spring", "f;0.1;0.0", "Minimizes distance from original location");
    createParameter("Anchor", "f;0.1;0.0", "Minimizes third power of distance from original location");
    createParameter("Iterations", "i;5;1", "Number of iterations");
    createParameter("Chillout", "i;5;0", "Number of finishing iterations");
}
bool SmoothDiameters::runMethod()
{
    Tree* mt = getCurrentTree();
    if(mt == 0) return false;
    float d1 = getFloatParameter("Alpha");
    float fa = getFloatParameter("Spring");
    float faaa = getFloatParameter("Anchor");
    int iterations = getIntParameter("Iterations");
    int chillout = getIntParameter("Chillout");
    SmoothDiameter smoothing(mt, d1, fa, faaa);
    Tree tree = smoothing.getResult(iterations, chillout);
    appendTree(tree, name);
    return true;
}


#include "../shared/treeextend.h"
CREATE_VKTOOL(ExtendBranches)
void ExtendBranches::initialize(void)
{
    createMethod("ExtendBranches", "Extends existing branches of selected tree");
    createParameter("MinCriterion", "f;0.5;0.0", "Minimum value to extend branch");
    createParameter("Sigma", "f;0.5;0.0", "Vesselness scale");
    createParameter("Normalize", "t;Yes;Yes;No");
    createParameter("Alpha1", "f;0.5;0.0", "Sato's vesselness computation parameter");
    createParameter("Alpha2", "f;2.0;0.0", "Sato's vesselness computation parameter");
    createParameter("MaxNodes", "i;100;1", "Maximum number of nodes added to a branch");
}
bool ExtendBranches::runMethod()
{
    Tree* mt = getCurrentTree();
    if(mt == 0) return false;
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;

    float mingoal = getFloatParameter("MinCriterion");
    float sigma = getFloatParameter("Sigma");
    std::string normalizes = getStringParameter("Normalize");
    bool normalize = (normalizes == "Yes");
    float alfa1 = getFloatParameter("Alpha1");
    float alfa2 = getFloatParameter("Alpha2");
    int maxnodes = getFloatParameter("MaxNodes");
    TreeExtend tool(mt, Mz2Itk<double>::getItk(mi), mingoal, maxnodes, sigma, normalize, alfa1, alfa2);
    Tree tree = tool.getResult();
    appendTree(tree, "extendedTree");
    return true;
}


CREATE_VKTOOL(DisjoinBranches)
void DisjoinBranches::initialize(void)
{
    createMethod("DisjoinBranches", "Disjoin branches of selected tree");
}
bool DisjoinBranches::runMethod()
{
    Tree* mt = getCurrentTree();
    if(mt == 0) return false;
    std::vector<Tree> ntrees;
    std::vector< std::string > nnames;
    std::vector< unsigned int > ncolors;

    unsigned int nb = mt->count();
    for(unsigned int n = 0; n < nb; n++)
    {
        Tree tree;
        std::vector<NodeIn3D> branch = mt->branch(n);
        if(branch.size() > 0)
        {
            tree.addBranch(branch);
            ntrees.push_back(tree);
            std::stringstream ss;
            ss << "branch" << n;
            nnames.push_back(ss.str());
            ncolors.push_back(0x00ff88);
        }
    }
    appendTrees(ntrees, nnames, ncolors);
    return true;
}


#include "../shared/tree.h"
CREATE_VKTOOL(JoinChecked)
void JoinChecked::initialize(void)
{
    createMethod("JoinChecked", "Connects checked trees and branches");
}
bool JoinChecked::runMethod()
{
    Tree* tree;
    Tree newtree;
    unsigned int i = 0;
    bool checked;
    bool add = false;
    while((tree = getTree(i, NULL, &checked)) != NULL)
    {
        if(checked)
        {
            unsigned int nb = tree->count();
            for(unsigned int b = 0; b < nb; b++)
            {
                std::vector<NodeIn3D> branch = tree->branch(b);
                newtree.addBranch(branch, false);
                add = true;
            }
        }
        i++;
    }
    if(add)
    {
        newtree.correctConnectivity();
        appendTree(newtree, name);
        return true;
    }
    return false;
}

#include "../shared/tree.h"
CREATE_VKTOOL(ConvertBranches)
void ConvertBranches::initialize(void)
{
    createMethod("ConvertBranches", "Changes storage format of branches between short and long");
    createParameter("Format", "t;Short;Short;Long");
}
bool ConvertBranches::runMethod()
{
    Tree* mt = getCurrentTree();
    if(mt == 0) return false;
    Tree tree = *(mt);
    std::string format = getStringParameter("Format");

    if(format == "Long")
    {
        if(tree.rebuildTreeLongBranches())
        {
            appendTree(tree, ("longBranchesFormat"), 0x77ff00);
        }
    }
    else
    {
        if(tree.rebuildTreeShortBranches())
        {
            appendTree(tree, ("shortBranchesFormat"), 0x77ff00);
        }
    }
    return true;
}


CREATE_VKTOOL(LevelSetChanVeseDense)
void LevelSetChanVeseDense::initialize(void)
{
    createMethod("LevelSetChanVeseDense", "");
    createParameter("Iterations", "i;50;1");
    createParameter("InitialDistance", "f;1.0;0.0");
    createParameter("Rms", "f;0.0;0.0");
    createParameter("Epsilon", "f;1.0;0.0");
    createParameter("CurvatureWeight", "f;0.0;0.0");
    createParameter("AreaWeight", "f;0.0;0.0");
    createParameter("ReinitializationWeight", "f;0.0;0.0");
    createParameter("VolumeWeight", "f;0.0;0.0");
    createParameter("Volume", "f;0.0;0.0");
    createParameter("Lambda1", "f;1.0;0.0");
    createParameter("Lambda2", "f;1.0;0.0");

}
bool LevelSetChanVeseDense::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;

    unsigned int nb_iteration = getIntParameter("Iterations");
    double initialDistance = getFloatParameter("InitialDistance");
    double rms = getFloatParameter("Rms");
    double epsilon = getFloatParameter("Epsilon");
    double curvature_weight = getFloatParameter("CurvatureWeight");
    double area_weight = getFloatParameter("AreaWeight");
    double reinitialization_weight = getFloatParameter("ReinitializationWeight");
    double volume_weight = getFloatParameter("VolumeWeight");
    double volume = getFloatParameter("Volume");
    double l1 = getFloatParameter("Lambda1");
    double l2 = getFloatParameter("Lambda2");

    Byte3DImage::IndexType fiducial;
    if(!getCurrentFiducial(&fiducial)) return false;

    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters2<float>::levelSetChanVeseDense(ii, fiducial,
                                                                                    nb_iteration,
                                                                                    initialDistance,
                                                                                    rms,
                                                                                    epsilon,
                                                                                    curvature_weight,
                                                                                    area_weight,
                                                                                    reinitialization_weight,
                                                                                    volume_weight,
                                                                                    volume,
                                                                                    l1,
                                                                                    l2);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);

    ImageWithInfo oui;
    oui.i = ni;
    oui.t = name;
    appendImage(oui, name);
    return true;
}



CREATE_VKTOOL(LevelSetChanVeseSparse)
void LevelSetChanVeseSparse::initialize(void)
{
    createMethod("LevelSetChanVeseSparse", "");
    createParameter("Iterations", "i;50;1");
    createParameter("InitialDistance", "f;1.0;0.0");
    createParameter("Rms", "f;0.0;0.0");
    createParameter("Epsilon", "f;1.0;0.0");
    createParameter("CurvatureWeight", "f;0.0;0.0");
    createParameter("AreaWeight", "f;0.0;0.0");
    createParameter("ReinitializationWeight", "f;0.0;0.0");
    createParameter("VolumeWeight", "f;0.0;0.0");
    createParameter("Volume", "f;0.0;0.0");
    createParameter("Lambda1", "f;1.0;0.0");
    createParameter("Lambda2", "f;1.0;0.0");

}
bool LevelSetChanVeseSparse::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;

    unsigned int nb_iteration = getIntParameter("Iterations");
    double initialDistance = getFloatParameter("InitialDistance");
    double rms = getFloatParameter("Rms");
    double epsilon = getFloatParameter("Epsilon");
    double curvature_weight = getFloatParameter("CurvatureWeight");
    double area_weight = getFloatParameter("AreaWeight");
    double reinitialization_weight = getFloatParameter("ReinitializationWeight");
    double volume_weight = getFloatParameter("VolumeWeight");
    double volume = getFloatParameter("Volume");
    double l1 = getFloatParameter("Lambda1");
    double l2 = getFloatParameter("Lambda2");

    Byte3DImage::IndexType fiducial;
    if(!getCurrentFiducial(&fiducial)) return false;

    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters2<float>::levelSetChanVeseSparse(ii, fiducial,
                                                                                    nb_iteration,
                                                                                    initialDistance,
                                                                                    rms,
                                                                                    epsilon,
                                                                                    curvature_weight,
                                                                                    area_weight,
                                                                                    reinitialization_weight,
                                                                                    volume_weight,
                                                                                    volume,
                                                                                    l1,
                                                                                    l2);
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);

    ImageWithInfo oui;
    oui.i = ni;
    oui.t = name;
    appendImage(oui, name);
    return true;
}


CREATE_VKTOOL(LevelSetCurves)
void LevelSetCurves::initialize(void)
{
    createMethod("LevelSetCurves", "");
    createParameter("Iterations", "i;800;1");
    createParameter("SmoothingIterations", "i;5;0");
    createParameter("InitialDistance", "f;1.0;0.0");
    createParameter("PropagationScaling", "f;1.0;0.0");
    createParameter("Sigma", "f;2.0;0.0");
    createParameter("Alpha", "f;1.0;0.0");
    createParameter("Beta", "f;1.0;0.0");
    createParameter("ThresholdLower", "f;-1000.0");
    createParameter("ThresholdUpper", "f;0.0;0.0");
    createParameter("ThresholdOutside", "f;0.0;0.0");
    createParameter("ThresholdInside", "f;255.0;0.0");
    createParameter("SigmoidMin", "f;0.0;0.0");
    createParameter("SigmoidMax", "f;1.0;0.0");
    createParameter("CurvatureScaling", "f;1.0;0.0");
    createParameter("AdvectionScaling", "f;1.0;0.0");
    createParameter("Rms", "f;0.02;0.0");
    createParameter("SmoothingTimeStep", "f;0.125;0.0");
    createParameter("SmoothingConductance", "f;9.0;0.0");
}
bool LevelSetCurves::runMethod()
{
    ImageWithInfo* mmi = getCurrentImage(); if(mmi == NULL) return false; Byte3DImage::Pointer mi = mmi->i;

    double initialDistance = getFloatParameter("InitialDistance");
    double propagationScaling = getFloatParameter("PropagationScaling");
    double sigma = getFloatParameter("Sigma");
    double alpha = getFloatParameter("Alpha");
    double beta = getFloatParameter("Beta");
    double thrLower = getFloatParameter("ThresholdLower");
    double thrUpper = getFloatParameter("ThresholdUpper");
    double thrOutside = getFloatParameter("ThresholdOutside");
    double thrInside = getFloatParameter("ThresholdInside");
    double sigMin = getFloatParameter("SigmoidMin");
    double sigMax = getFloatParameter("SigmoidMax");
    double curvSc = getFloatParameter("CurvatureScaling");
    double advSc = getFloatParameter("AdvectionScaling");
    double rms = getFloatParameter("Rms");
    unsigned int iterations = getIntParameter("Iterations");
    double smoTimeStep = getFloatParameter("SmoothingTimeStep");
    unsigned int smoIterations = getIntParameter("SmoothingIterations");
    double smoConductance = getFloatParameter("SmoothingConductance");

    Byte3DImage::IndexType fiducial;
    if(!getCurrentFiducial(&fiducial)) return false;

    Mz2Itk<float>::Image3DType::Pointer ii = Mz2Itk<float>::getItk(mi);
    Mz2Itk<float>::Image3DType::Pointer oi = ImageFilters2<float>::levelSetCurves(ii, fiducial,
                                                                                  initialDistance,
                                                                                  propagationScaling,
                                                                                  sigma, alpha, beta,
                                                                                  thrLower, thrUpper, thrOutside, thrInside,
                                                                                  sigMin, sigMax, curvSc, advSc, rms,
                                                                                  iterations, smoTimeStep, smoIterations, smoConductance
                                                                                  );
    Byte3DImage::Pointer ni = Mz2Itk<float>::getMZNormalized(oi);

    ImageWithInfo oui;
    oui.i = ni;
    oui.t = name;
    appendImage(oui, name);
    return true;
}

#endif // VKTOOLS_H
